﻿//---------------------------------------------------------------------------
//
//   Thicken
//
//          Copyright(C) 2013-2017, tetraface Inc.
//
//     Sample of Select plug-in for thickening selected faces.
//     Created DLL must be installed in "Plugins\Select" directory.
//
//    　選択した面に厚みをつける選択プラグインのサンプル。
//    　作成したDLLは"Plugins\Select"フォルダに入れる必要があります。
//
//
//---------------------------------------------------------------------------

#ifdef WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <Shlwapi.h>
#endif
#include "MQPlugin.h"
#include "MQBasePlugin.h"
#include "MQSetting.h"
#include "MQ3DLib.h"
#include "MQWidget.h"
#include "Language.h"
#if __APPLE__
#include <CoreFoundation/CFBundle.h>
#include <CoreFoundation/CFByteOrder.h>
#endif

#ifdef _WIN32
static wchar_t s_DllPath[_MAX_PATH];
#endif



#ifdef WIN32
#define FILEPATH_DELIMITER_CHAR L'\\'
#define FILEPATH_DELIMITER_L    L"\\"
#define FILEPATH_DELIMITER      "\\"
#else
#define FILEPATH_DELIMITER_CHAR L'/'
#define FILEPATH_DELIMITER_L    L"/"
#define FILEPATH_DELIMITER      "/"
#endif

inline static bool isPathSeparator(wchar_t character)
{
	return (character == L'\\' || character == L'/');
}

inline static wchar_t createPathSeparator(const std::wstring& path)
{
	for(const wchar_t *ptr = path.c_str() + path.length(); ptr > path.c_str(); ){
		--ptr;
		if(isPathSeparator(*ptr)){
			return *ptr;
		}
	}
	return FILEPATH_DELIMITER_CHAR; // return default separator
}

std::wstring extractDirectory(const std::wstring& src_path)
{
#ifdef WIN32
	wchar_t buf[MAX_PATH];
	wcscpy_s(buf, src_path.c_str());
	::PathRemoveFileSpecW(buf);
	return std::wstring(buf);
#else
	size_t len = src_path.length();
	for(const wchar_t *ptr = src_path.c_str() + len; ptr > src_path.c_str(); ){
		--ptr;
		if(isPathSeparator(*ptr)){
			std::wstring ret = src_path.substr(0, ptr - src_path.c_str());
			if(ret.length() > 0 && !isPathSeparator(ret[ret.length()-1])){
				ret += createPathSeparator(ret);
			}
			return ret;
		}
	}
	return std::wstring();
#endif
}

std::wstring combinePath(const std::wstring& base_dir, const std::wstring& cat_dir)
{
#ifdef _WIN32
	wchar_t buf[MAX_PATH];
	::PathCombineW(buf, base_dir.c_str(), cat_dir.c_str());
	return std::wstring(buf);
#endif
#if __APPLE__
	std::wstring ret;
	if(base_dir.length() > 0){
		ret += base_dir;
		if(base_dir[base_dir.length()-1] != L'\\' && base_dir[base_dir.length()-1] != L'/'){
			ret += L"/";
		}
	}
	ret += cat_dir;
	std::wstring::size_type p = ret.find_first_of(L'\\');
	while(p != std::wstring::npos){
		ret[p] = L'/';
		p = ret.find(L'\\', p+1);
	}
	return ret;
#endif
}


#ifdef _WIN32
//---------------------------------------------------------------------------
//  DllMain
//---------------------------------------------------------------------------
BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved
					 )
{
	switch(ul_reason_for_call){
	case DLL_PROCESS_ATTACH:
		::GetModuleFileName((HMODULE)hModule, s_DllPath, _MAX_PATH);
		break;
	}

	// プラグインとしては特に必要な処理はないので、何もせずにTRUEを返す
	// There is nothing to do for a plugin. So just return TRUE.
    return TRUE;
}
#endif
#if __APPLE__
static std::wstring GetResourceDir()
{
	// Get a resource directory.
	CFBundleRef bundleRef = CFBundleGetBundleWithIdentifier(CFSTR("tetraface.Thicken"));
	if(bundleRef == NULL) return std::wstring();
	
	CFURLRef resourceUrl = CFBundleCopyResourcesDirectoryURL(bundleRef);
	char path[PATH_MAX+100];
	Boolean ret = CFURLGetFileSystemRepresentation(resourceUrl, TRUE, (UInt8*)path, PATH_MAX+100);
	CFRelease(resourceUrl);
	if(!ret) return std::wstring();
	
	return MQEncoding::Utf8ToWide(path);
}
#endif

//---------------------------------------------------------------------------
//  class ThickenPlugin
//---------------------------------------------------------------------------
class ThickenPlugin : public MQSelectPlugin
{
public:
	ThickenPlugin();
	~ThickenPlugin();

	void GetPlugInID(DWORD *Product, DWORD *ID) override;
	const char *GetPlugInName(void) override;
	const char *EnumString(int index) override;
	BOOL Execute(int index, MQDocument doc) override;

	MLanguage m_Language;

private:
	BOOL Thicken(MQDocument doc);
	void LoadResource();
};


MQBasePlugin *GetPluginClass()
{
	static ThickenPlugin plugin;
	return &plugin;
}


ThickenPlugin::ThickenPlugin()
{
}

ThickenPlugin::~ThickenPlugin()
{
}

//---------------------------------------------------------------------------
//  GetPlugInID
//    プラグインIDを返す。
//    この関数は起動時に呼び出される。
//---------------------------------------------------------------------------
void ThickenPlugin::GetPlugInID(DWORD *Product, DWORD *ID)
{
	// プロダクト名(制作者名)とIDを、全部で64bitの値として返す
	// 値は他と重複しないようなランダムなもので良い
	*Product = 0x56A31D20;
	*ID      = 0xA5E0FE68;
}

//---------------------------------------------------------------------------
//  GetPlugInName
//    プラグイン名を返す。
//    この関数は[プラグインについて]表示時に呼び出される。
//---------------------------------------------------------------------------
const char *ThickenPlugin::GetPlugInName(void)
{
	// プラグイン名
	return "Thicken       Copyright(C) 2013-2017, tetraface Inc.";
}

//---------------------------------------------------------------------------
//  EnumString
//    ポップアップメニューに表示される文字列を返す。
//    この関数は起動時に呼び出される。
//---------------------------------------------------------------------------
const char *ThickenPlugin::EnumString(int index)
{
	LoadResource();

	static std::string str("Thicken");
	std::wstring wstr = m_Language.Search("Title");
	if(wstr.length() > 0){
		str = MQEncoding::WideToAnsi(wstr.c_str());
	}
	
	switch(index){
	case 0:
		return str.c_str();
	}
	return NULL;
}

//---------------------------------------------------------------------------
//  Execute
//    メニューから選択されたときに呼び出される。
//---------------------------------------------------------------------------
BOOL ThickenPlugin::Execute(int index, MQDocument doc)
{
	// Load language resources
	LoadResource();

	switch(index){
	case 0: return Thicken(doc);
	}
	return FALSE;
}

void ThickenPlugin::LoadResource()
{
	if(m_Language.GetLanguage().length() == 0){
		std::wstring lang = GetSettingValue(MQSETTINGVALUE_LANGUAGE);
#ifdef WIN32
		std::wstring dir = extractDirectory(s_DllPath);
#else
		std::wstring dir = GetResourceDir();
#endif
		std::wstring path = combinePath(dir, L"Thicken.resource.xml");
		m_Language.Load(lang, path);
	}
}


class ThickenDialog : public MQDialog
{
public:
	ThickenDialog(MQWindowBase& parent, ThickenPlugin *plugin);
	~ThickenDialog();

	BOOL ThicknessChanged(MQWidgetBase *sender, MQDocument doc);
	BOOL DirectionChanged(MQWidgetBase *sender, MQDocument doc);

	MQDoubleSpinBox *thickness;
	MQRadioButton *dir_radio[3];
	MQCheckBox *curmat_check;
};

ThickenDialog::ThickenDialog(MQWindowBase& parent, ThickenPlugin *plugin) : MQDialog(parent)
{
	SetTitle(plugin->m_Language.Search("Title"));

	MQFrame *mainFrame = CreateVerticalFrame(this);
	MQFrame *frame;

	// Thickness
	frame = CreateHorizontalFrame(mainFrame);
	CreateLabel(frame, plugin->m_Language.Search("Thickness"));
	thickness = CreateDoubleSpinBox(frame);
	thickness->SetMin(1e-8);
	thickness->SetMax(1e+8);
	thickness->SetPosition(5);
	thickness->SetDisplayUnit(plugin->GetDisplayUnit());
	thickness->AddChangingEvent(this, &ThickenDialog::ThicknessChanged);

	// Direction
	frame = CreateHorizontalFrame(mainFrame);
	CreateLabel(frame, plugin->m_Language.Search("Direction"));
	dir_radio[0] = CreateRadioButton(frame, plugin->m_Language.Search("DirFront"));
	dir_radio[1] = CreateRadioButton(frame, plugin->m_Language.Search("DirCenter"));
	dir_radio[2] = CreateRadioButton(frame, plugin->m_Language.Search("DirBack"));
	dir_radio[0]->AddChangedEvent(this, &ThickenDialog::DirectionChanged);
	dir_radio[1]->AddChangedEvent(this, &ThickenDialog::DirectionChanged);
	dir_radio[2]->AddChangedEvent(this, &ThickenDialog::DirectionChanged);

	// Current material
	frame = CreateHorizontalFrame(mainFrame);
	curmat_check = CreateCheckBox(frame, plugin->m_Language.Search("SideCurMat"));

	MQFrame *bottomFrame = CreateHorizontalFrame(mainFrame);
	MQButton *okbtn = CreateButton(bottomFrame, plugin->GetResourceString("OK"));
	okbtn->SetDefault(true);
	okbtn->SetModalResult(MQDialog::DIALOG_OK);
	okbtn->SetHintSizeRateX(4.0);
	okbtn->SetFillBeforeRate(1.0);

	MQButton *cancelbtn = CreateButton(bottomFrame, plugin->GetResourceString("Cancel"));
	cancelbtn->SetCancel(true);
	cancelbtn->SetModalResult(MQDialog::DIALOG_CANCEL);
	cancelbtn->SetHintSizeRateX(4.0);
	cancelbtn->SetFillAfterRate(1.0);
	MQButton::ReorderOKButton(bottomFrame, okbtn, cancelbtn);
}

ThickenDialog::~ThickenDialog()
{
}

BOOL ThickenDialog::ThicknessChanged(MQWidgetBase *sender, MQDocument doc)
{
	return FALSE;
}

BOOL ThickenDialog::DirectionChanged(MQWidgetBase *sender, MQDocument doc)
{
	for(int i=0; i<3; i++){
		if(dir_radio[i] != sender){
			dir_radio[i]->SetChecked(false);
		}
	}
	return FALSE;
}


BOOL ThickenPlugin::Thicken(MQDocument doc)
{
	{
		int num_obj = doc->GetObjectCount();
		bool selected = false;
		for(int oi=0; oi<num_obj && !selected; oi++)
		{
			MQObject obj = doc->GetObject(oi);
			if(obj == NULL) continue;

			int num_face = obj->GetFaceCount();
			int num_vert = obj->GetVertexCount();

			// Find target vertices in selected faces.
			for(int fi=0; fi<num_face; fi++){
				int num = obj->GetFacePointCount(fi);
				if(num >= 3 && doc->IsSelectFace(oi, fi)){
					selected = true;
					break;
				}
			}
		}
		if(!selected){
			MQWindow mainwin = MQWindow::GetMainWindow();
			MQDialog::MessageWarningBox(mainwin, GetResourceString("FaceErr"), GetResourceString("Error"));
			return FALSE;
		}
	}

	float thickness = 10.0f;
	int dir = 0;
	bool side_curmat = false;

	// Load the settings.
	MQSetting *setting = OpenSetting();
	if(setting != NULL){
		setting->Load("Thickness", thickness, thickness);
		setting->Load("Direction", dir, dir);
		setting->Load("SideCurMat", side_curmat, side_curmat);
		CloseSetting(setting);
	}

	// ダイアログを表示
	// Show a dialog.
	MQWindow mainwin = MQWindow::GetMainWindow();
	ThickenDialog dialog(mainwin, this);
	dialog.thickness->SetPosition(thickness);
	if(dir > 0){
		dialog.dir_radio[0]->SetChecked(true);
	}else if(dir < 0){
		dialog.dir_radio[2]->SetChecked(true);
	}else{
		dialog.dir_radio[1]->SetChecked(true);
	}
	dialog.curmat_check->SetChecked(side_curmat);
	
	if(dialog.Execute() != MQDialog::DIALOG_OK){
		return FALSE;
	}

	// Get parameters from the dialog.
	thickness = (float)dialog.thickness->GetPosition();
	if(dialog.dir_radio[0]->GetChecked()) dir = 1;
	if(dialog.dir_radio[1]->GetChecked()) dir = 0;
	if(dialog.dir_radio[2]->GetChecked()) dir = -1;
	side_curmat = dialog.curmat_check->GetChecked();
	
	// Save the settings.
	setting = OpenSetting();
	if(setting != NULL){
		setting->Save("Thickness", thickness);
		setting->Save("Direction", dir);
		setting->Save("SideCurMat", side_curmat);
		CloseSetting(setting);
	}

	MQWaitCursorChanger wait_cursor(this);
	int curmat = doc->GetCurrentMaterialIndex();
	int num_obj = doc->GetObjectCount();
	for(int oi=0; oi<num_obj; oi++)
	{
		MQObject obj = doc->GetObject(oi);
		if(obj == NULL) continue;

		int num_face = obj->GetFaceCount();
		int num_vert = obj->GetVertexCount();
		try
		{
			std::vector<bool> face_sel(num_face, false);
			std::vector<bool> vert_sel(num_vert, false);
			std::vector<int> vi, nvi, rel, rvi;

			// Find target vertices in selected faces.
			int num_select_face = 0;
			for(int fi=0; fi<num_face; fi++){
				int num = obj->GetFacePointCount(fi);
				if(num >= 3 && doc->IsSelectFace(oi, fi)){
					face_sel[fi] = true;
					vi.resize(num);
					obj->GetFacePointArray(fi, &(*vi.begin()));
				
					for(int j=0; j<num; j++){
						vert_sel[vi[j]] = true;
					}
					num_select_face++;
				}
			}
			if(num_select_face == 0) continue;

			MQObjEdge edge(obj);
			MQObjNormal normal(obj);

			float offset1, offset2;
			if(dir > 0){
				offset1 = thickness;
				offset2 = 0;
			}else if(dir < 0){
				offset1 = 0;
				offset2 = -thickness;
			}else{
				offset1 = thickness*0.5f;
				offset2 = -thickness*0.5f;
			}

			// Create new vertices.
			std::vector<int> new_vi1(num_vert, -1);
			std::vector<int> new_vi2(num_vert, -1);
			for(int vi=0; vi<num_vert; vi++){
				if(vert_sel[vi]){
					int num_rel = obj->GetVertexRelatedFaces(vi, NULL);
					rel.resize(num_rel);
					obj->GetVertexRelatedFaces(vi, &(*rel.begin()));

					MQPoint vn(0,0,0);
					for(int rfi=0; rfi<num_rel; rfi++){
						int num = obj->GetFacePointCount(rel[rfi]);
						rvi.resize(num);
						obj->GetFacePointArray(rel[rfi], &(*rvi.begin()));
						for(int j=0; j<num; j++){
							if(rvi[j] == vi){
								vn += normal.Get(rel[rfi], j);
								break;
							}
						}
					}
					vn = Normalize(vn);

					MQPoint p = obj->GetVertex(vi);
					if(offset1 != 0){
						new_vi1[vi] = obj->AddVertex(p + vn * offset1);
						if(new_vi1[vi] == -1)
							throw std::bad_alloc();
					}else{
						new_vi1[vi] = vi;
					}
					if(offset2 != 0){
						new_vi2[vi] = obj->AddVertex(p + vn * offset2);
						if(new_vi2[vi] == -1)
							throw std::bad_alloc();
					}else{
						new_vi2[vi] = vi;
					}
				}
			}

			// Create new faces.
			std::vector<MQCoordinate> uv, nuv;
			for(int fi=0; fi<num_face; fi++){
				if(face_sel[fi]){
					int num = obj->GetFacePointCount(fi);
					vi.resize(num);
					obj->GetFacePointArray(fi, &(*vi.begin()));

					nvi.resize(num);
					int nfi;
					int mi = obj->GetFaceMaterial(fi);
					uv.resize(num);
					obj->GetFaceCoordinateArray(fi, &(*uv.begin()));

					// Front face
					for(int j=0; j<num; j++){
						nvi[j] = new_vi1[vi[j]];
					}
					nfi = obj->AddFace(num, &(*nvi.begin()));
					if(nfi == -1)
						throw std::bad_alloc();
					obj->SetFaceMaterial(nfi, mi);
					obj->SetFaceCoordinateArray(nfi, &(*uv.begin()));

					// Back face
					for(int j=0; j<num; j++){
						nvi[j] = new_vi2[vi[num-1-j]];
					}
					nfi = obj->AddFace(num, &(*nvi.begin()));
					if(nfi == -1)
						throw std::bad_alloc();
					obj->SetFaceMaterial(nfi, mi);
					nuv = uv;
					std::reverse(nuv.begin(), nuv.end());
					obj->SetFaceCoordinateArray(nfi, &(*nuv.begin()));

					// Side faces
					for(int j=0; j<num; j++){
						int pf, pl;
						// Check if the neighbor faces is selected.
						bool side = false;
						if(edge.getPair(fi, j, pf, pl)){
							if(!face_sel[pf]){
								side = true;
							}
						}else{
							side = true;
						}

						// Create a side face if not selected for not exist.
						if(side){
							int svi[4];
							svi[0] = new_vi1[vi[(j+1)%num]];
							svi[1] = new_vi1[vi[j]];
							svi[2] = new_vi2[vi[j]];
							svi[3] = new_vi2[vi[(j+1)%num]];
							nfi = obj->AddFace(4, svi);
							if(nfi == -1)
								throw std::bad_alloc();
							if(side_curmat){
								obj->SetFaceMaterial(nfi, curmat);
							}else{
								obj->SetFaceMaterial(nfi, mi);
							}
							nuv.resize(4);
							nuv[0] = nuv[3] = uv[(j+1)%num];
							nuv[1] = nuv[2] = uv[j];
							obj->SetFaceCoordinateArray(nfi, &(*nuv.begin()));
						}
					}
				}
			}

			// Delete the old face.
			for(int fi=0; fi<num_face; fi++){
				if(face_sel[fi]){
					obj->DeleteFace(fi);
				}
			}
		}
		catch(std::bad_alloc&){
			wait_cursor.End();
			MQWindow mainwin = MQWindow::GetMainWindow();
			MQDialog::MessageWarningBox(mainwin, GetResourceString("MemoryErr"), GetResourceString("Error"));
			for(int vi=num_vert; vi<obj->GetVertexCount(); vi++){
				obj->DeleteVertex(vi);
			}
			for(int fi=num_face; fi<obj->GetFaceCount(); fi++){
				obj->DeleteFace(fi);
			}
			return TRUE;
		}
	}

	return TRUE;
}

